/*
 * Copyright 1993-2007 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:   
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and 
 * international Copyright laws.  Users and possessors of this source code 
 * are hereby granted a nonexclusive, royalty-free license to use this code 
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE 
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR 
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH 
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF 
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL, 
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS 
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE 
 * OR PERFORMANCE OF THIS SOURCE CODE.  
 *
 * U.S. Government End Users.   This source code is a "commercial item" as 
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of 
 * "commercial computer  software"  and "commercial computer software 
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995) 
 * and is provided to the U.S. Government only as a commercial end item.  
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through 
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the 
 * source code with only those rights set forth herein. 
 *
 * Any use of this source code in individual and commercial software must 
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

/*
 * This sample implements full-range histogram calculation
 * of arbitrary-sized 8-bit data array
 */



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>



#define ATOMICS
//Comment this line and add '-arch sm_11' to the command line to enable atomic operations.
#undef ATOMICS


////////////////////////////////////////////////////////////////////////////////
// Reference CPU histogram
////////////////////////////////////////////////////////////////////////////////
extern "C" void histogram256CPU(
    unsigned int *h_Result,
    unsigned int *h_Data,
    int dataN
);



////////////////////////////////////////////////////////////////////////////////
// GPU histogram
////////////////////////////////////////////////////////////////////////////////
#include <histogram256_kernel.cu>



////////////////////////////////////////////////////////////////////////////////
// Data configuration
////////////////////////////////////////////////////////////////////////////////
//Input data size
const int          DATA_N = 100000000;
const int       DATA_SIZE = DATA_N * sizeof(unsigned char);


int main(int argc, char **argv){
    unsigned char
        *h_Data;

    unsigned int
        *h_ResultCPU, *h_ResultGPU;

    unsigned char
        *d_Data;

    int i, sum, delta;
    double timerValue;

    unsigned int hTimer;


    CUT_DEVICE_INIT();
    CUT_SAFE_CALL(cutCreateTimer(&hTimer));

    printf("Initializing data...\n");
        printf("...allocating CPU memory.\n");
        h_Data      = (unsigned char *)malloc(DATA_SIZE);
        h_ResultCPU = (unsigned int *)malloc(HISTOGRAM_SIZE);
        h_ResultGPU = (unsigned int *)malloc(HISTOGRAM_SIZE);

        printf("...generating input data\n");
        srand((unsigned int)time(NULL));
        for(i = 0; i < DATA_N; i++)
            h_Data[i] = rand() % 256;

        printf("...allocating GPU memory and copying input data\n");
        CUDA_SAFE_CALL( cudaMalloc((void **)&d_Data,   DATA_SIZE ) );
        CUDA_SAFE_CALL( cudaMemcpy(d_Data, h_Data, DATA_SIZE, cudaMemcpyHostToDevice) );

    printf("Running GPU histogram...\n");
        initHistogram256();
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
        histogram256GPU(h_ResultGPU, (unsigned int *) d_Data, DATA_N);
        CUDA_SAFE_CALL( cudaThreadSynchronize() );
        CUT_SAFE_CALL(cutStopTimer(hTimer));
        timerValue = cutGetTimerValue(hTimer);
        closeHistogram256();
    printf("histogram256gpu() time : %f msec //%f MB/sec\n", timerValue, DATA_SIZE / (1048576.0 * timerValue * 0.001));


    printf("Comparing the results...\n");
        printf("...histogramCPU()\n");
        CUT_SAFE_CALL( cutResetTimer(hTimer) );
        CUT_SAFE_CALL( cutStartTimer(hTimer) );
        histogram256CPU(
            h_ResultCPU,
            (unsigned int *)h_Data,
            DATA_N / 4
        );
        CUT_SAFE_CALL(cutStopTimer(hTimer));
        timerValue = cutGetTimerValue(hTimer);
    printf("histogramCPU() time : %f msec //%f MB/sec\n", timerValue, DATA_SIZE / (1048576.0 * timerValue * 0.001));

        sum   = 0;
        delta = 0;
        for(i = 0; i < BIN_COUNT; i++){
            sum   += h_ResultGPU[i];
            delta += abs((int)h_ResultGPU[i] - (int)h_ResultCPU[i]);
        }

    printf("Total sum of histogram elements: %i\n", sum);
    printf("Sum of absolute differences: %i\n", delta);
    printf((delta == 0) ? "TEST PASSED\n" : "TEST FAILED\n");

    printf("Shutting down...\n");
        CUT_SAFE_CALL(cutDeleteTimer(hTimer));
        CUDA_SAFE_CALL( cudaFree(d_Data) );
        free(h_ResultGPU);
        free(h_ResultCPU);
        free(h_Data);

    CUT_EXIT(argc, argv);
}
